from gitea_client.auth import Authentication, Token, UsernamePassword
from gitea_client.entities import GiteaUser, GiteaRepo, GiteaBranch, GiteaCommit, GiteaOrg, GiteaTeam
from gitea_client.interface import GiteaApi, ApiFailure, NetworkFailure
from gitea_client.updates import GiteaUserUpdate, GiteaHookUpdate



import base64
import json
import unittest

import responses
from future.moves.urllib.parse import parse_qs

import gitea_client
import gitea_client._implementation.http_utils as http_utils


class GiteaClientInterfaceTest(unittest.TestCase):
    def setUp(self):
        self.api_endpoint = "https://www.example.com/api/v1/"
        self.base_url = "https://www.example.com/"
        self.client = gitea_client.GiteaApi(self.base_url)
        self.repo_json_str = """{
                "id": 27,
                "owner": {
                    "id": 1,
                    "username": "unknwon",
                    "full_name": "",
                    "email": "u@gitea.io",
                    "avatar_url": "/avatars/1"
                  },
                  "name": "Hello-World",
                  "full_name": "unknwon/Hello-World",
                  "description": "Some description",
                  "private": false,
                  "fork": false,
                  "parent": null,
                  "default_branch": "master",
                  "empty": false,
                  "size": 42,
                  "html_url": "http://localhost:3000/unknwon/Hello-World",
                  "clone_url": "http://localhost:3000/unknwon/hello-world.git",
                  "ssh_url": "jiahuachen@localhost:unknwon/hello-world.git",
                  "permissions": {
                    "admin": true,
                    "push": true,
                    "pull": true
                  }
                }"""
        self.repos_list_json_str = """[{
                "id": 27,
                "owner": {
                    "id": 1,
                    "username": "unknwon",
                    "full_name": "",
                    "email": "u@gitea.io",
                    "avatar_url": "/avatars/1"
                  },
                  "name": "Hello-World",
                  "full_name": "unknwon/Hello-World",
                  "description": "Some description",
                  "private": false,
                  "fork": false,
                  "parent": null,
                  "default_branch": "master",
                  "empty": false,
                  "size": 42,
                  "html_url": "http://localhost:3000/unknwon/Hello-World",
                  "clone_url": "http://localhost:3000/unknwon/hello-world.git",
                  "ssh_url": "jiahuachen@localhost:unknwon/hello-world.git",
                  "permissions": {
                    "admin": true,
                    "push": true,
                    "pull": true
                  }
                },{
                "id": 28,
                "owner": {
                    "id": 1,
                    "username": "unknwon",
                    "full_name": "",
                    "email": "u@gitea.io",
                    "avatar_url": "/avatars/1"
                  },
                  "name": "Hello-World-Again",
                  "full_name": "unknwon/Hello-World-Again",
                  "private": false,
                  "fork": false,
                  "parent": null,
                  "default_branch": "master",
                  "empty": false,
                  "size": 42,
                  "html_url": "http://localhost:3000/unknwon/Hello-World-Again",
                  "clone_url": "http://localhost:3000/unknwon/hello-world-again.git",
                  "ssh_url": "jiahuachen@localhost:unknwon/hello-world-again.git",
                  "permissions": {
                    "admin": true,
                    "push": true,
                    "pull": true
                  }
                }]"""
        self.branch_json_str = """{
                        "name": "master",
                        "commit": {
                            "id": "c17825309a0d52201e78a19f49948bcc89e52488",
                            "message": "migrated to RC0.2",
                            "url": "Not implemented",
                            "author": {
                                "name": "Joel Lonbeck",
                                "email": "joel@neutrinographics.com",
                                "username": "joel"
                            },
                            "committer": {
                                "name": "Joel Lonbeck",
                                "email": "joel@neutrinographics.com",
                                "username": "joel"
                            },
                            "verification": {
                                "verified": false,
                                "reason": "gpg.error.not_signed_commit",
                                "signature": "",
                                "payload": ""
                            },
                            "timestamp": "2017-05-17T21:11:25Z"
                            }
                        }"""
        self.branches_list_json_str = """[{
                        "name": "master",
                        "commit": {
                            "id": "c17825309a0d52201e78a19f49948bcc89e52488",
                            "message": "migrated to RC0.2",
                            "url": "Not implemented",
                            "author": {
                                "name": "Joel Lonbeck",
                                "email": "joel@neutrinographics.com",
                                "username": "joel"
                            },
                            "committer": {
                                "name": "Joel Lonbeck",
                                "email": "joel@neutrinographics.com",
                                "username": "joel"
                            },
                            "verification": {
                                "verified": false,
                                "reason": "gpg.error.not_signed_commit",
                                "signature": "",
                                "payload": ""
                            },
                            "timestamp": "2017-05-17T21:11:25Z"
                            }
                        },{
                        "name": "develop",
                        "commit": {
                            "id": "r03kd7cjr9a0d52201e78a19f49948bcc89e52488",
                            "message": "another branch",
                            "url": "Not implemented",
                            "author": {
                                "name": "Joel Lonbeck",
                                "email": "joel@neutrinographics.com",
                                "username": "joel"
                            },
                            "committer": {
                                "name": "Joel Lonbeck",
                                "email": "joel@neutrinographics.com",
                                "username": "joel"
                            },
                            "verification": {
                                "verified": false,
                                "reason": "gpg.error.not_signed_commit",
                                "signature": "",
                                "payload": ""
                            },
                            "timestamp": "2017-05-17T21:11:25Z"
                            }
                        }]"""
        self.user_json_str = """{
                  "id": 1,
                  "username": "unknwon",
                  "full_name": "",
                  "email": "u@gitea.io",
                  "avatar_url": "/avatars/1"
                }"""
        self.token_json_str = """{
                  "name": "new token",
                  "sha1": "mytoken"
                }"""
        self.username_password = gitea_client.UsernamePassword(
            "auth_username", "password")
        self.hook_json_str = """{
                "id": 4,
                "type": "gitea",
                "config": {
                  "content_type": "json",
                  "url": "http://test.io/hook"
                },
                "events": [
                  "create",
                  "push",
                  "issues"
                ],
                "active": false,
                "updated_at": "2017-03-31T12:42:58Z",
                "created_at": "2017-03-31T12:42:58Z"
              }"""
        self.hooks_list_json_str = """[
              {
                "id": 4,
                "type": "gitea",
                "config": {
                  "content_type": "json",
                  "url": "http://test.io/hook"
                },
                "events": [
                  "create",
                  "push",
                  "issues"
                ],
                "active": false,
                "updated_at": "2017-03-31T12:42:58Z",
                "created_at": "2017-03-31T12:42:58Z"
              },
              {
                "id": 3,
                "type": "gitea",
                "config": {
                  "content_type": "json",
                  "url": "http://192.168.201.1:8080/hook22/"
                },
                "events": [
                  "issue_comment"
                ],
                "active": true,
                "updated_at": "2017-03-31T12:47:56Z",
                "created_at": "2017-03-31T12:42:54Z"
              }
            ]"""
        self.org_json_str = """{
              "id": 7,
              "username": "gitea2",
              "full_name": "Gitea2",
              "avatar_url": "/avatars/7",
              "description": "Gitea is a painless self-hosted Git Service.",
              "website": "https://gitea.io",
              "location": "USA"
            }"""
        self.team_json_str = """{
              "id": 12,
              "name": "new-team",
              "description": "A new team created by API",
              "permission": "write"
            }"""
        self.deploy_key_json_str = """{
              "id": 1,
              "key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDUbmwBOG5vI8qNCztby5LDc9ozwTuwsqf+1fpuHjT9iQ2Lu9nlKHQJcPSgdrYAcc+88K6o74ayhTAjfajKxkIHnbzZFjidoVZSQDhX5qvl93jvY/Uz390qky0sweW+fspm8pRJL+ofE3QEN5AXAuycq1tgsRT32XC+Ta82Xyv8b3xW+pWbsZzYCzUsZXDe/xWxg1rndXh2BIrmcYf9BMiv9ZJIojJXfuLCeRXl550tDzaMFC0rQ/T5pZjs/lQemtg92MnxnEDi5nhuvDwM4Q8eqCTOXc4BCE7iyIHv+B7rx+0x99ytMh5BSIIGyWTfgTot/AjGVm5aRKJSRFgPBm9N comment with whitespace",
              "url": "http://localhost:3000/api/v1/repos/unknwon/project_x/keys/1",
              "title": "local",
              "created_at": "2015-11-18T15:05:43-05:00",
              "read_only": true
            }"""
        self.deploy_key_json_list = """[{
              "id": 1,
              "key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDUbmwBOG5vI8qNCztby5LDc9ozwTuwsqf+1fpuHjT9iQ2Lu9nlKHQJcPSgdrYAcc+88K6o74ayhTAjfajKxkIHnbzZFjidoVZSQDhX5qvl93jvY/Uz390qky0sweW+fspm8pRJL+ofE3QEN5AXAuycq1tgsRT32XC+Ta82Xyv8b3xW+pWbsZzYCzUsZXDe/xWxg1rndXh2BIrmcYf9BMiv9ZJIojJXfuLCeRXl550tDzaMFC0rQ/T5pZjs/lQemtg92MnxnEDi5nhuvDwM4Q8eqCTOXc4BCE7iyIHv+B7rx+0x99ytMh5BSIIGyWTfgTot/AjGVm5aRKJSRFgPBm9N comment with whitespace",
              "url": "http://localhost:3000/api/v1/repos/unknwon/project_x/keys/1",
              "title": "local",
              "created_at": "2015-11-18T15:05:43-05:00",
              "read_only": true
            },{
              "id": 2,
              "key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDUbmwBOG5vI8qNCztby5LDc9ozwTuwsqf+1fpuHjT9iQ2Lu9nlKHQJcPSgdrYAcc+88K6o74ayhTAjfajKxkIHnbzZFjidoVZSQDhX5qvl93jvY/Uz390qky0sweW+fspm8pRJL+ofE3QEN5AXAuycq1tgsRT32XC+Ta82Xyv8b3xW+pWbsZzYCzUsZXDe/xWxg1rndXh2BIrmcYf9BMiv9ZJIojJXfuLCeRXl550tDzaMFC0rQ/T5pZjs/lQemtg92MnxnEDi5nhuvDwM4Q8eqCTOXc4BCE7iyIHv+B7rx+0x99ytMh5BSIIGyWTfgTot/AjGVm5aRKJSRFgPBm9N comment with whitespace",
              "url": "http://localhost:3000/api/v1/repos/unknwon/project_x/keys/1",
              "title": "local",
              "created_at": "2015-11-18T15:05:43-05:00",
              "read_only": true
            }]"""
        self.expected_repo = gitea_client.GiteaRepo.from_json(json.loads(self.repo_json_str))
        self.expected_branch = gitea_client.GiteaBranch.from_json(json.loads(self.branch_json_str))
        self.expected_user = gitea_client.GiteaUser.from_json(json.loads(self.user_json_str))
        self.expected_hook = gitea_client.GiteaRepo.Hook.from_json(json.loads(self.hook_json_str))
        self.expected_org = gitea_client.GiteaOrg.from_json(json.loads(self.org_json_str))
        self.expected_team = gitea_client.GiteaTeam.from_json(json.loads(self.team_json_str))
        self.expected_key = gitea_client.GiteaRepo.DeployKey.from_json(json.loads(self.deploy_key_json_str))
        self.token = gitea_client.Token.from_json(json.loads(self.token_json_str))

    @responses.activate
    def test_create_repo1(self):
        uri = self.path("/user/repos")
        responses.add(responses.POST, uri, body=self.repo_json_str)
        repo = self.client.create_repo(self.token, "AGreatRepo")
        self.assert_repos_equal(repo, self.expected_repo)
        self.assertEqual(len(responses.calls), 1)
        call = responses.calls[0]
        self.assertEqual(call.request.url, self.path_with_token(uri))

    @responses.activate
    def test_repo_exists1(self):
        uri1 = self.path("/repos/username/repo1")
        uri2 = self.path("/repos/username/repo2")
        responses.add(responses.GET, uri1, body=self.repo_json_str, status=200)
        responses.add(responses.GET, uri2, status=404)
        self.assertTrue(self.client.repo_exists(self.token, "username", "repo1"))
        self.assertFalse(self.client.repo_exists(self.token, "username", "repo2"))
        self.assertEqual(len(responses.calls), 2)
        first_call = responses.calls[0]
        self.assertEqual(first_call.request.url, self.path_with_token(uri1))
        last_call = responses.calls[1]
        self.assertEqual(last_call.request.url, self.path_with_token(uri2))

    @responses.activate
    def test_get_repo1(self):
        uri1 = self.path("/repos/username/repo1")
        uri2 = self.path("/repos/username/repo2")
        responses.add(responses.GET, uri1, body=self.repo_json_str, status=200)
        responses.add(responses.GET, uri2, status=404)
        repo = self.client.get_repo(self.token, "username", "repo1")
        self.assert_repos_equal(repo, self.expected_repo)
        self.assertRaises(gitea_client.ApiFailure, self.client.get_repo,
                          self.token, "username", "repo2")
        self.assertEqual(len(responses.calls), 2)
        first_call = responses.calls[0]
        self.assertEqual(first_call.request.url, self.path_with_token(uri1))
        last_call = responses.calls[1]
        self.assertEqual(last_call.request.url, self.path_with_token(uri2))

    @responses.activate
    def test_get_user_repos(self):
        uri = self.path("/users/username/repos")
        responses.add(responses.GET, uri, body=self.repos_list_json_str, status=200)
        repos = self.client.get_user_repos(self.token, "username")
        self.assertEqual(len(repos), 2)
        self.assert_repos_equal(repos[0], self.expected_repo)

    @responses.activate
    def test_delete_repo1(self):
        uri1 = self.path("/repos/username/repo1")
        uri2 = self.path("/repos/otherusername/repo2")
        responses.add(responses.DELETE, uri1, status=204)
        responses.add(responses.DELETE, uri2, status=401)
        self.client.delete_repo(self.token, "username", "repo1")
        self.assertRaises(gitea_client.ApiFailure, self.client.delete_repo,
                          self.token, "otherusername", "repo2")
        self.assertEqual(len(responses.calls), 2)
        first_call = responses.calls[0]
        self.assertEqual(first_call.request.url, self.path_with_token(uri1))
        last_call = responses.calls[1]
        self.assertEqual(last_call.request.url, self.path_with_token(uri2))

    @responses.activate
    def test_get_branch(self):
        uri = self.path("/repos/username/repo/branches/branch")
        responses.add(responses.GET, uri, body=self.branch_json_str, status=200)
        branch = self.client.get_branch(self.token, "username", "repo", "branch")
        self.assert_branches_equal(branch, self.expected_branch)

    @responses.activate
    def test_get_branches(self):
        uri = self.path("/repos/username/repo/branches")
        responses.add(responses.GET, uri, body=self.branches_list_json_str, status=200)
        branches = self.client.get_branches(self.token, "username", "repo")
        self.assertEqual(len(branches), 2)
        self.assert_branches_equal(branches[0], self.expected_branch)

    @responses.activate
    def test_create_user1(self):
        uri = self.path("/admin/users")
        responses.add(responses.POST, uri, body=self.user_json_str, status=201)
        user = self.client.create_user(self.token, login_name="loginname",
                                       username="username", email="user@example.com",
                                       password="password")
        self.assert_users_equals(user, self.expected_user)
        self.assertEqual(len(responses.calls), 1)
        call = responses.calls[0]
        self.assertEqual(call.request.url, self.path_with_token(uri))

    @responses.activate
    def test_user_exists1(self):
        uri1 = self.path("/users/username1")
        uri2 = self.path("/users/username2")
        responses.add(responses.GET, uri1, body=self.user_json_str, status=200)
        responses.add(responses.GET, uri2, status=404)
        self.assertTrue(self.client.user_exists("username1"))
        self.assertFalse(self.client.user_exists("username2"))
        self.assertEqual(len(responses.calls), 2)
        first_call = responses.calls[0]
        self.assertEqual(first_call.request.url, uri1)
        last_call = responses.calls[1]
        self.assertEqual(last_call.request.url, uri2)

    @responses.activate
    def test_search_users(self):
        uri = self.path("/users/search")

        def callback(request):
            url = request.url
            index = url.find("?")
            self.assertTrue(index >= 0)  # assert a ? was actually found
            self.assertEqual(url[:index], uri)
            data = self.data_of_query(url[index + 1:])
            self.assertEqual(data["q"], "keyword")
            self.assertEqual(int(data["limit"]), 4)
            return 200, {}, "{\"data\": [%s], \"ok\": true}" % self.user_json_str

        responses.add_callback(responses.GET, uri, callback=callback)
        users = self.client.search_users("keyword", limit=4)
        self.assertEqual(len(users), 1)
        self.assert_users_equals(users[0], self.expected_user)

    @responses.activate
    def test_get_user1(self):
        uri1 = self.path("/users/username1")
        uri2 = self.path("/users/username2")
        responses.add(responses.GET, uri1, body=self.user_json_str, status=200)
        responses.add(responses.GET, uri2, status=404)
        user = self.client.get_user(None, "username1")
        self.assert_users_equals(user, self.expected_user)
        try:
            self.client.get_user(None, "username2")
            raise AssertionError("Call to get_user did not raise an exception")
        except gitea_client.ApiFailure as exc:
            self.assertIsNotNone(exc.message)
            self.assertEqual(exc.status_code, 404)
        self.assertEqual(len(responses.calls), 2)
        first_call = responses.calls[0]
        self.assertEqual(first_call.request.url, uri1)
        last_call = responses.calls[1]
        self.assertEqual(last_call.request.url, uri2)

    @responses.activate
    def test_update_user1(self):
        update = gitea_client.GiteaUserUpdate.Builder("loginname", "user@example.com") \
            .set_full_name("Example User") \
            .set_password("Password") \
            .set_website("mywebsite.net") \
            .set_active(True) \
            .set_admin(False) \
            .set_allow_git_hook(False) \
            .build()

        def callback(request):
            data = json.loads(request.body.decode('utf8'))
            self.assertEqual(data["login_name"], "loginname")
            self.assertEqual(data["full_name"], "Example User")
            self.assertEqual(data["email"], "user@example.com")
            self.assertEqual(data["password"], "Password")
            self.assertEqual(data["website"], "mywebsite.net")
            self.assertRegexpMatches(str(data["active"]), r"[tT]rue")
            self.assertRegexpMatches(str(data["admin"]), r"[fF]alse")
            self.assertRegexpMatches(str(data["allow_git_hook"]), r"[fF]alse")
            self.assertNotIn("id", data)
            self.assertNotIn("location", data)
            self.assertNotIn("allow_import_local", data)
            return 200, {}, self.user_json_str

        uri = self.path("/admin/users/username")
        responses.add_callback(responses.PATCH, uri, callback=callback)
        user = self.client.update_user(self.token, "username", update)
        self.assert_users_equals(user, self.expected_user)

    @responses.activate
    def test_delete_user1(self):
        uri1 = self.path("/admin/users/username1")
        uri2 = self.path("/admin/users/username2")
        responses.add(responses.DELETE, uri1, status=204)
        responses.add(responses.DELETE, uri2, status=401)
        self.client.delete_user(self.username_password, "username1")
        self.assertRaises(gitea_client.ApiFailure, self.client.delete_user,
                          self.username_password, "username2")
        self.assertEqual(len(responses.calls), 2)
        first_call = responses.calls[0]
        self.assertEqual(first_call.request.url, uri1)
        self.check_for_basic_auth(first_call.request)
        last_call = responses.calls[1]
        self.assertEqual(last_call.request.url, uri2)
        self.check_for_basic_auth(last_call.request)

    @responses.activate
    def test_valid_authentication1(self):
        uri = self.path("/user")
        valid_token = gitea_client.Token("a_valid_token")
        invalid_token = gitea_client.Token("an_invalid_token")

        def callback(request):
            if request.url == self.path_with_token(uri, valid_token):
                return 200, {}, self.user_json_str
            elif request.url == self.path_with_token(uri, invalid_token):
                return 401, {}, ""
            else:
                self.fail("Unexpected URL: {}".format(request.url))

        responses.add_callback(responses.GET, uri, callback=callback)
        self.assertTrue(self.client.valid_authentication(valid_token))
        self.assertFalse(self.client.valid_authentication(invalid_token))

    @responses.activate
    def test_authenticated_user(self):
        uri = self.path("/user")
        responses.add(responses.GET, uri, body=self.user_json_str, status=200)
        user = self.client.authenticated_user(self.token)
        self.assert_users_equals(user, self.expected_user)

    @responses.activate
    def test_ensure_token(self):
        uri = self.path("/users/{}/tokens".format(self.username_password.username))
        responses.add(responses.GET, uri, body="[]", status=200)
        responses.add(responses.POST, uri, body=self.token_json_str, status=200)
        responses.add(responses.GET, uri, body="[" + self.token_json_str + "]", status=200)
        token = self.client.ensure_token(self.username_password, self.token.name, self.username_password.username)
        self.assert_tokens_equals(token, self.token)
        token = self.client.ensure_token(self.username_password, self.token.name, self.username_password.username)
        self.assert_tokens_equals(token, self.token)

    @responses.activate
    def test_ensure_auth_token(self):
        uri = self.path("/user")
        responses.add(responses.GET, uri, body=self.user_json_str, status=200)
        uri = self.path("/users/{}/tokens".format(self.expected_user.username))
        responses.add(responses.GET, uri, body="[]", status=200)
        responses.add(responses.POST, uri, body=self.token_json_str, status=200)
        tokens = self.client.get_tokens(self.username_password)
        self.assertEqual(tokens, [])
        tokeninfo = self.client.create_token(self.username_password, self.token.name)
        self.assert_tokens_equals(tokeninfo, self.token)
        token = self.client.ensure_token(self.username_password, self.token.name)
        self.assert_tokens_equals(token, self.token)

    @responses.activate
    def test_create_hook1(self):
        uri = self.path("/repos/username/repo1/hooks")
        responses.add(responses.POST, uri, body=self.hook_json_str)
        hook = self.client.create_hook(self.token,
                                       repo_name="repo1",
                                       hook_type="gitea",
                                       config={
                                           "content_type": "json2",
                                           "url": "http://test.io/hook"
                                       },
                                       events=["create", "push", "issues"],
                                       active=False,
                                       organization="username")
        self.assert_hooks_equals(hook, self.expected_hook)
        self.assertEqual(len(responses.calls), 1)
        call = responses.calls[0]
        self.assertEqual(call.request.url, self.path_with_token(uri))

    @responses.activate
    def test_update_hook1(self):
        update = gitea_client.GiteaHookUpdate.Builder() \
            .set_events(["issues_comments"]) \
            .set_config({"url": "http://newurl.com/hook"}) \
            .set_active(True) \
            .build()

        def callback(request):
            data = json.loads(request.body.decode('utf8'))
            self.assertEqual(data["config"]["url"], "http://newurl.com/hook")
            self.assertEqual(data["events"], ['issues_comments'])
            self.assertEqual(data["active"], True)
            return 200, {}, self.hook_json_str

        uri = self.path("/repos/username/repo1/hooks/4")
        responses.add_callback(responses.PATCH, uri, callback=callback)
        hook = self.client.update_hook(self.token, "repo1", 4, update, organization="username")
        self.assert_hooks_equals(hook, self.expected_hook)

    @responses.activate
    def test_list_hooks(self):
        uri = self.path("/repos/username/repo1/hooks")
        responses.add(responses.GET, uri, body=self.hooks_list_json_str, status=200)
        hooks = self.client.get_repo_hooks(self.token, "username", "repo1")
        self.assertEqual(len(hooks), 2)
        self.assert_hooks_equals(hooks[0], self.expected_hook)

    @responses.activate
    def test_delete_hook(self):
        uri = self.path("/repos/username/repo1/hooks/4")
        responses.add(responses.DELETE, uri, status=204)
        hook = self.client.delete_hook(self.token, "username", "repo1", 4)
        self.assertEqual(hook, None)

    @responses.activate
    def test_create_organization(self):
        uri = self.path("/admin/users/username/orgs")
        responses.add(responses.POST, uri, body=self.org_json_str)
        org = self.client.create_organization(self.token,
                                              owner_name="username",
                                              org_name="gitea2",
                                              full_name="Gitea2",
                                              description="Gitea is a painless self-hosted Git Service.",
                                              website="https://gitea.io",
                                              location="USA")
        self.assert_org_equals(org, self.expected_org)
        self.assertEqual(len(responses.calls), 1)
        call = responses.calls[0]
        self.assertEqual(call.request.url, self.path_with_token(uri))

    @responses.activate
    def test_create_organization_team(self):
        uri = self.path("/admin/orgs/username/teams")
        responses.add(responses.POST, uri, body=self.team_json_str)
        team = self.client.create_organization_team(self.token,
                                                    org_name="username",
                                                    name="new-team",
                                                    description="A new team created by API",
                                                    permission="write")
        self.assert_team_equals(team, self.expected_team)
        self.assertEqual(len(responses.calls), 1)
        call = responses.calls[0]
        self.assertEqual(call.request.url, self.path_with_token(uri))

    @responses.activate
    def test_add_team_membership(self):
        uri = self.path("/admin/teams/team/members/username")
        responses.add(responses.PUT, uri, status=204)
        resp = self.client.add_team_membership(self.token, "team", "username")
        self.assertEqual(resp, None)

    @responses.activate
    def test_remove_team_membership(self):
        uri = self.path("/admin/teams/team/members/username")
        responses.add(responses.DELETE, uri, status=204)
        resp = self.client.remove_team_membership(self.token, "team", "username")
        self.assertEqual(resp, None)

    @responses.activate
    def test_add_repo_to_team(self):
        uri = self.path("/admin/teams/test_team/repos/repo_name")
        responses.add(responses.PUT, uri, status=204)
        resp = self.client.add_repo_to_team(self.token, "test_team", "repo_name")
        self.assertEqual(resp, None)

    @responses.activate
    def test_remove_repo_from_team(self):
        uri = self.path("/admin/teams/test_team/repos/repo_name")
        responses.add(responses.DELETE, uri, status=204)
        resp = self.client.remove_repo_from_team(self.token, "test_team", "repo_name")
        self.assertEqual(resp, None)

    @responses.activate
    def test_list_deploy_keys(self):
        uri = self.path("/repos/username/repo1/keys")
        responses.add(responses.GET, uri, body=self.deploy_key_json_list, status=200)
        keys = self.client.list_deploy_keys(self.token, "username", "repo1")
        self.assertEqual(len(keys), 2)
        self.assert_keys_equals(keys[0], self.expected_key)

    @responses.activate
    def test_delete_deploy_key(self):
        uri = self.path("/repos/username/repo1/keys/1")
        responses.add(responses.DELETE, uri, status=204)
        self.client.delete_deploy_key(self.token, "username", "repo1", 1)
        call = responses.calls[0]
        self.assertEqual(call.request.url, self.path_with_token(uri))

    @responses.activate
    def test_get_deploy_key(self):
        uri = self.path("/repos/username/repo1/keys/1")
        responses.add(responses.GET, uri, body=self.deploy_key_json_str)
        key = self.client.get_deploy_key(self.token, "username", "repo1", 1)
        self.assert_keys_equals(key, self.expected_key)

    @responses.activate
    def test_add_deploy_key(self):
        uri = self.path("/repos/username/repo1/keys")
        responses.add(responses.POST, uri, body=self.deploy_key_json_str)
        key_title = "My key title"
        key_content = "My key content"
        key = self.client.add_deploy_key(self.token, "username", "repo1", key_title, key_content)
        self.assert_keys_equals(key, self.expected_key)

    # helper methods

    @staticmethod
    def data_of_query(query):
        return {k: v[0] for (k, v) in parse_qs(query).items()}

    def path(self, relative):
        return http_utils.append_url(self.api_endpoint, relative)

    def path_with_token(self, path, token=None):
        token = self.token if token is None else token
        return "{p}?token={t}".format(p=path, t=token.token)

    def check_for_basic_auth(self, request):
        auth = "{u}:{p}".format(u=self.username_password.username,
                                p=self.username_password.password)
        b64 = base64.b64encode(auth.encode()).decode()
        self.assertEqual(request.headers["Authorization"], "Basic {}".format(b64))

    def assert_repos_equal(self, repo, expected):
        self.assertEqual(repo.id, expected.id)
        self.assert_users_equals(repo.owner, expected.owner)
        self.assertEqual(repo.full_name, expected.full_name)
        self.assertEqual(repo.private, expected.private)
        self.assertEqual(repo.fork, expected.fork)
        self.assertEqual(repo.parent, expected.parent)
        self.assertEqual(repo.default_branch, expected.default_branch)
        self.assertEqual(repo.size, expected.size)
        self.assertEqual(repo.empty, expected.empty)
        self.assertEqual(repo.urls.html_url, expected.urls.html_url)
        self.assertEqual(repo.urls.clone_url, expected.urls.clone_url)
        self.assertEqual(repo.urls.ssh_url, expected.urls.ssh_url)
        self.assertEqual(repo.permissions.admin, expected.permissions.admin)
        self.assertEqual(repo.permissions.push, expected.permissions.push)
        self.assertEqual(repo.permissions.pull, expected.permissions.pull)

    def assert_branches_equal(self, branch, expected):
        self.assertEqual(branch.name, expected.name)
        self.assert_commits_equal(branch.commit, expected.commit)

    def assert_commits_equal(self, commit, expected):
        self.assertEqual(commit.id, expected.id)
        self.assertEqual(commit.message, expected.message)
        self.assertEqual(commit.url, expected.url)
        self.assertEqual(commit.timestamp, expected.timestamp)

    def assert_users_equals(self, user, expected):
        self.assertEqual(user.id, expected.id)
        self.assertEqual(user.username, expected.username)
        self.assertEqual(user.full_name, expected.full_name)
        self.assertEqual(user.email, expected.email)
        self.assertEqual(user.avatar_url, expected.avatar_url)

    def assert_tokens_equals(self, token, expected):
        self.assertEqual(token.name, expected.name)
        self.assertEqual(token.token, expected.token)

    def assert_hooks_equals(self, hook, expected):
        self.assertEqual(hook.id, expected.id)
        self.assertEqual(hook.type, expected.type)
        self.assertEqual(hook.events, expected.events)
        self.assertEqual(hook.config, expected.config)
        self.assertEqual(hook.active, expected.active)

    def assert_org_equals(self, org, expected):
        self.assertEqual(org.id, expected.id)
        self.assertEqual(org.username, expected.username)
        self.assertEqual(org.full_name, expected.full_name)
        self.assertEqual(org.avatar_url, expected.avatar_url)
        self.assertEqual(org.description, expected.description)
        self.assertEqual(org.website, expected.website)
        self.assertEqual(org.location, expected.location)

    def assert_team_equals(self, team, expected):
        self.assertEqual(team.id, expected.id)
        self.assertEqual(team.name, expected.name)
        self.assertEqual(team.description, expected.description)
        self.assertEqual(team.permission, expected.permission)

    def assert_keys_equals(self, key, expected):
        self.assertEqual(key.id, expected.id)
        self.assertEqual(key.title, expected.title)
        self.assertEqual(key.url, expected.url)
        self.assertEqual(key.key, expected.key)
        self.assertEqual(key.read_only, expected.read_only)
        self.assertEqual(key.created_at, expected.created_at)


if __name__ == "__main__":
    unittest.main()



import requests

from gitea_client._implementation.http_utils import RelativeHttpRequestor, append_url
from gitea_client.auth import Token
from gitea_client.entities import GiteaUser, GiteaRepo, GiteaBranch, GiteaOrg, GiteaTeam


class GiteaApi(object):
    """
    A Gitea client, serving as a wrapper around the Gitea HTTP API.
    """

    def __init__(self, base_url, session=None):
        """
        :param str base_url: the URL of the Gitea server to communicate with. Should be given
                             with the https protocol
        :param requests.Session session: a ``requests`` session instance
        """
        api_base = append_url(base_url, "/api/v1/")
        self._requestor = RelativeHttpRequestor(api_base, session=session)

    def valid_authentication(self, auth):
        """
        Returns whether ``auth`` is valid

        :param auth.Authentication auth: authentication object
        :return: whether the provided authentication is valid
        :rtype: bool
        :raises NetworkFailure: if there is an error communicating with the server
        """
        return self._get("/user", auth=auth).ok

    def authenticated_user(self, auth):
        """
        Returns the user authenticated by ``auth``

        :param auth.Authentication auth: authentication for user to retrieve

        :return: user authenticated by the provided authentication
        :rtype: GiteaUser
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        response = self.get("/user", auth=auth)
        return GiteaUser.from_json(response.json())

    def get_tokens(self, auth, username=None):
        """
        Returns the tokens owned by the specified user. If no user is specified,
        uses the user authenticated by ``auth``.

        :param auth.Authentication auth: authentication for user to retrieve.
                                         Must be a username-password authentication,
                                         due to a restriction of the Gitea API
        :param str username: username of owner of tokens

        :return: list of tokens
        :rtype: List[Token]
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        if username is None:
            username = self.authenticated_user(auth).username
        response = self.get("/users/{u}/tokens".format(u=username), auth=auth)
        return [Token.from_json(o) for o in response.json()]

    def create_token(self, auth, name, username=None):
        """
        Creates a new token with the specified name for the specified user.
        If no user is specified, uses user authenticated by ``auth``.

        :param auth.Authentication auth: authentication for user to retrieve.
                                         Must be a username-password authentication,
                                         due to a restriction of the Gitea API
        :param str name: name of new token
        :param str username: username of owner of new token

        :return: new token representation
        :rtype: Token
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        if username is None:
            username = self.authenticated_user(auth).username
        data = {"name": name}
        response = self.post("/users/{u}/tokens".format(u=username), auth=auth, data=data)
        return Token.from_json(response.json())

    def ensure_token(self, auth, name, username=None):
        """
        Ensures the existence of a token with the specified name for the
        specified user. Creates a new token if none exists. If no user is
        specified, uses user authenticated by ``auth``.

        :param auth.Authentication auth: authentication for user to retrieve.
                                         Must be a username-password authentication,
                                         due to a restriction of the Gitea API
        :param str name: name of new token
        :param str username: username of owner of new token

        :return: token representation
        :rtype: Token
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        if username is None:
            username = self.authenticated_user(auth).username
        tokens = [token for token in self.get_tokens(auth, username) if token.name == name]
        if len(tokens) > 0:
            return tokens[0]
        return self.create_token(auth, name, username)

    def create_repo(self, auth, name, description=None, private=False, auto_init=False,
                    gitignore_templates=None, license_template=None, readme_template=None,
                    organization=None):
        """
        Creates a new repository, and returns the created repository.

        :param auth.Authentication auth: authentication object
        :param str name: name of the repository to create
        :param str description: description of the repository to create
        :param bool private: whether the create repository should be private
        :param bool auto_init: whether the created repository should be auto-initialized with an initial commit
        :param list[str] gitignore_templates: collection of ``.gitignore`` templates to apply
        :param str license_template: license template to apply
        :param str readme_template: README template to apply
        :param str organization: organization under which repository is created
        :return: a representation of the created repository
        :rtype: GiteaRepo
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        gitignores = None if gitignore_templates is None \
            else ",".join(gitignore_templates)
        data = {
            "name": name,
            "description": description,
            "private": private,
            "auto_init": auto_init,
            "gitignores": gitignores,
            "license": license_template,
            "readme": readme_template
        }
        data = {k: v for (k, v) in data.items() if v is not None}
        url = "/org/{0}/repos".format(organization) if organization else "/user/repos"
        response = self.post(url, auth=auth, data=data)
        return GiteaRepo.from_json(response.json())

    def repo_exists(self, auth, username, repo_name):
        """
        Returns whether a repository with name ``repo_name`` owned by the user with username ``username`` exists.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :param str repo_name: name of repository
        :return: whether the repository exists
        :rtype: bool
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/repos/{u}/{r}".format(u=username, r=repo_name)
        return self._get(path, auth=auth).ok

    def get_repo(self, auth, username, repo_name):
        """
        Returns a the repository with name ``repo_name`` owned by
        the user with username ``username``.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :param str repo_name: name of repository
        :return: a representation of the retrieved repository
        :rtype: GiteaRepo
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/repos/{u}/{r}".format(u=username, r=repo_name)
        response = self.get(path, auth=auth)
        return GiteaRepo.from_json(response.json())

    def get_user_repos(self, auth, username):
        """
        Returns the repositories  owned by
        the user with username ``username``.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :return: a list of repositories
        :rtype: List[GiteaRepo]
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/users/{u}/repos".format(u=username)
        response = self.get(path, auth=auth)
        return [GiteaRepo.from_json(repo_json) for repo_json in response.json()]

    def get_branch(self, auth, username, repo_name, branch_name):
        """
        Returns the branch with name ``branch_name`` in the repository with name ``repo_name``
        owned by the user with username ``username``.

        :param auth.Authentication auth: authentication object 
        :param str username: username of owner of repository containing the branch
        :param str repo_name: name of the repository with the branch
        :param str branch_name: name of the branch to return
        :return: a branch
        :rtype: GiteaBranch
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/repos/{u}/{r}/branches/{b}".format(u=username, r=repo_name, b=branch_name)
        response = self.get(path, auth=auth)
        return GiteaBranch.from_json(response.json())

    def get_branches(self, auth, username, repo_name):
        """
        Returns the branches in the repository with name ``repo_name`` owned by the user
        with username ``username``.

        :param auth.Authentication auth: authentication object 
        :param str username: username of owner of repository containing the branch
        :param str repo_name: name of the repository with the branch
        :return: a list of branches
        :rtype: List[GiteaBranch]
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/repos/{u}/{r}/branches".format(u=username, r=repo_name)
        response = self.get(path, auth=auth)
        return [GiteaBranch.from_json(branch_json) for branch_json in response.json()]

    def delete_repo(self, auth, username, repo_name):
        """
        Deletes the repository with name ``repo_name`` owned by the user with username ``username``.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository to delete
        :param str repo_name: name of repository to delete
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/repos/{u}/{r}".format(u=username, r=repo_name)
        self.delete(path, auth=auth)

    def migrate_repo(self, auth, clone_addr,
                     uid, repo_name, auth_username=None, auth_password=None,
                     mirror=False, private=False, description=None):
        """
        Migrate a repository from another Git hosting source for the authenticated user.

        :param auth.Authentication auth: authentication object
        :param str clone_addr: Remote Git address (HTTP/HTTPS URL or local path)
        :param int uid: user ID of repository owner
        :param str repo_name: Repository name
        :param bool mirror: Repository will be a mirror. Default is false
        :param bool private: Repository will be private. Default is false
        :param str description: Repository description
        :return: a representation of the migrated repository
        :rtype: GiteaRepo
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        # "auth_username": auth_username,
        # "auth_password": auth_password,

        data = {
            "clone_addr": clone_addr,
            "uid": uid,
            "repo_name": repo_name,
            "mirror": mirror,
            "private": private,
            "description": description,
        }
        data = {k: v for (k, v) in data.items() if v is not None}
        url = "/repos/migrate"
        response = self.post(url, auth=auth, data=data)
        return GiteaRepo.from_json(response.json())

    def create_user(self, auth, login_name, username, email, password, send_notify=False):
        """
        Creates a new user, and returns the created user.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str login_name: login name for created user
        :param str username: username for created user
        :param str email: email address for created user
        :param str password: password for created user
        :param bool send_notify: whether a notification email should be sent upon creation
        :return: a representation of the created user
        :rtype: GiteaUser
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        # Since the source_id parameter was not well-documented at the time this method was
        # written, force the default value
        data = {
            "login_name": login_name,
            "username": username,
            "email": email,
            "password": password,
            "send_notify": send_notify
        }
        response = self.post("/admin/users", auth=auth, data=data)
        return GiteaUser.from_json(response.json())

    def user_exists(self, username):
        """
        Returns whether a user with username ``username`` exists.

        :param str username: username of user
        :return: whether a user with the specified username exists
        :rtype: bool
        :raises NetworkFailure: if there is an error communicating with the server
        :return:
        """
        path = "/users/{}".format(username)
        return self._get(path).ok

    def search_users(self, username_keyword, limit=10):
        """
        Searches for users whose username matches ``username_keyword``, and returns
        a list of matched users.

        :param str username_keyword: keyword to search with
        :param int limit: maximum number of returned users
        :return: a list of matched users
        :rtype: List[GiteaUser]
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        params = {"q": username_keyword, "limit": limit}
        response = self.get("/users/search", params=params)
        return [GiteaUser.from_json(user_json) for user_json in response.json()["data"]]

    def get_user(self, auth, username):
        """
        Returns a representing the user with username ``username``.

        :param auth.Authentication auth: authentication object, can be ``None``
        :param str username: username of user to get
        :return: the retrieved user
        :rtype: GiteaUser
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/users/{}".format(username)
        response = self.get(path, auth=auth)
        return GiteaUser.from_json(response.json())

    def update_user(self, auth, username, update):
        """
        Updates the user with username ``username`` according to ``update``.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str username: username of user to update
        :param GiteaUserUpdate update: a ``GiteaUserUpdate`` object describing the requested update
        :return: the updated user
        :rtype: GiteaUser
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/admin/users/{}".format(username)
        response = self.patch(path, auth=auth, data=update.as_dict())
        return GiteaUser.from_json(response.json())

    def delete_user(self, auth, username):
        """
        Deletes the user with username ``username``. Should only be called if the
        to-be-deleted user has no repositories.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str username: username of user to delete
        """
        path = "/admin/users/{}".format(username)
        self.delete(path, auth=auth)

    def get_repo_hooks(self, auth, username, repo_name):
        """
        Returns all hooks of repository with name ``repo_name`` owned by
        the user with username ``username``.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :param str repo_name: name of repository
        :return: a list of hooks for the specified repository
        :rtype: List[GiteaRepo.Hooks]
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/repos/{u}/{r}/hooks".format(u=username, r=repo_name)
        response = self.get(path, auth=auth)
        return [GiteaRepo.Hook.from_json(hook) for hook in response.json()]

    def create_hook(self, auth, repo_name, hook_type, config, events=None, organization=None, active=False):
        """
        Creates a new hook, and returns the created hook.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str repo_name: the name of the repo for which we create the hook
        :param str hook_type: The type of webhook, either "gitea", "gitea" or "slack"
        :param dict config: Settings for this hook (possible keys are
                            ``"url"``, ``"content_type"``, ``"secret"``)
        :param list events: Determines what events the hook is triggered for. Default: ["push"]
        :param str organization: Organization of the repo
        :param bool active: Determines whether the hook is actually triggered on pushes. Default is false
        :return: a representation of the created hook
        :rtype: GiteaRepo.Hook
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        if events is None:
            events = ["push"]  # default value is mutable, so assign inside body

        data = {
            "type": hook_type,
            "config": config,
            "events": events,
            "active": active
        }

        url = "/repos/{o}/{r}/hooks".format(o=organization, r=repo_name) if organization is not None \
            else "/repos/{r}/hooks".format(r=repo_name)
        response = self.post(url, auth=auth, data=data)
        return GiteaRepo.Hook.from_json(response.json())

    def update_hook(self, auth, repo_name, hook_id, update, organization=None):
        """
        Updates hook with id ``hook_id`` according to ``update``.

        :param auth.Authentication auth: authentication object
        :param str repo_name: repo of the hook to update
        :param int hook_id: id of the hook to update
        :param GiteaHookUpdate update: a ``GiteaHookUpdate`` object describing the requested update
        :param str organization: name of associated organization, if applicable
        :return: the updated hook
        :rtype: GiteaRepo.Hook
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        if organization is not None:
            path = "/repos/{o}/{r}/hooks/{i}".format(o=organization, r=repo_name, i=hook_id)
        else:
            path = "/repos/{r}/hooks/{i}".format(r=repo_name, i=hook_id)
        response = self._patch(path, auth=auth, data=update.as_dict())
        return GiteaRepo.Hook.from_json(response.json())

    def delete_hook(self, auth, username, repo_name, hook_id):
        """
        Deletes the hook with id ``hook_id`` for repo with name ``repo_name``
        owned by the user with username ``username``.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :param str repo_name: name of repository of hook to delete
        :param int hook_id: id of hook to delete
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        path = "/repos/{u}/{r}/hooks/{i}".format(u=username, r=repo_name, i=hook_id)
        self.delete(path, auth=auth)

    def create_organization(self, auth, owner_name, org_name, full_name=None, description=None,
                            website=None, location=None):
        """
        Creates a new organization, and returns the created organization.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str owner_name: Username of organization owner
        :param str org_name: Organization name
        :param str full_name: Full name of organization 
        :param str description: Description of the organization
        :param str website: Official website
        :param str location: Organization location
        :return: a representation of the created organization
        :rtype: GiteaOrg
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        data = {
            "username": org_name,
            "full_name": full_name,
            "description": description,
            "website": website,
            "location": location
        }

        url = "/admin/users/{u}/orgs".format(u=owner_name)
        response = self.post(url, auth=auth, data=data)
        return GiteaOrg.from_json(response.json())

    def create_organization_team(self, auth, org_name, name, description=None, permission="read"):
        """
        Creates a new team of the organization.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str org_name: Organization user name
        :param str name: Full name of the team
        :param str description: Description of the team
        :param str permission: Team permission, can be read, write or admin, default is read
        :return: a representation of the created team
        :rtype: GiteaTeam
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        data = {
            "name": name,
            "description": description,
            "permission": permission
        }

        url = "/admin/orgs/{o}/teams".format(o=org_name)
        response = self.post(url, auth=auth, data=data)
        return GiteaTeam.from_json(response.json())

    def add_team_membership(self, auth, team_id, username):
        """
        Add user to team.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str team_id: Team's id
        :param str username: Username of the user to be added to team
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        url = "/admin/teams/{t}/members/{u}".format(t=team_id, u=username)
        self.put(url, auth=auth)

    def remove_team_membership(self, auth, team_id, username):
        """
        Remove user from team.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str team_id: Team's id
        :param str username: Username of the user to be removed from the team
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        url = "/admin/teams/{t}/members/{u}".format(t=team_id, u=username)
        self.delete(url, auth=auth)

    def add_repo_to_team(self, auth, team_id, repo_name):
        """
        Add or update repo from team.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str team_id: Team's id
        :param str repo_name: Name of the repo to be added to the team
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        url = "/admin/teams/{t}/repos/{r}".format(t=team_id, r=repo_name)
        self.put(url, auth=auth)

    def remove_repo_from_team(self, auth, team_id, repo_name):
        """
        Remove repo from team.

        :param auth.Authentication auth: authentication object, must be admin-level
        :param str team_id: Team's id
        :param str repo_name: Name of the repo to be removed from the team
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        url = "/admin/teams/{t}/repos/{r}".format(t=team_id, r=repo_name)
        self.delete(url, auth=auth)

    def list_deploy_keys(self, auth, username, repo_name):
        """
        List deploy keys for the specified repo.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :param str repo_name: the name of the repo
        :return: a list of deploy keys for the repo
        :rtype: List[GiteaRepo.DeployKey]
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        response = self.get("/repos/{u}/{r}/keys".format(u=username, r=repo_name), auth=auth)
        return [GiteaRepo.DeployKey.from_json(key_json) for key_json in response.json()]

    def get_deploy_key(self, auth, username, repo_name, key_id):
        """
        Get a deploy key for the specified repo.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :param str repo_name: the name of the repo
        :param int key_id: the id of the key
        :return: the deploy key
        :rtype: GiteaRepo.DeployKey
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        response = self.get("/repos/{u}/{r}/keys/{k}".format(u=username, r=repo_name, k=key_id), auth=auth)
        return GiteaRepo.DeployKey.from_json(response.json())

    def add_deploy_key(self, auth, username, repo_name, title, key_content):
        """
        Add a deploy key to the specified repo.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :param str repo_name: the name of the repo
        :param str title: title of the key to add
        :param str key_content: content of the key to add
        :return: a representation of the added deploy key
        :rtype: GiteaRepo.DeployKey
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        data = {
            "title": title,
            "key": key_content
        }
        response = self.post("/repos/{u}/{r}/keys".format(u=username, r=repo_name), auth=auth, data=data)
        return GiteaRepo.DeployKey.from_json(response.json())

    def delete_deploy_key(self, auth, username, repo_name, key_id):
        """
        Remove deploy key for the specified repo.

        :param auth.Authentication auth: authentication object
        :param str username: username of owner of repository
        :param str repo_name: the name of the repo
        :param int key_id: the id of the key
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        self.delete("/repos/{u}/{r}/keys/{k}".format(u=username, r=repo_name, k=key_id), auth=auth)

    # Helper methods

    def _delete(self, path, auth=None, **kwargs):
        if auth is not None:
            auth.update_kwargs(kwargs)
        try:
            return self._requestor.delete(path, **kwargs)
        except requests.RequestException as exc:
            raise NetworkFailure(exc)

    def delete(self, path, auth=None, **kwargs):
        """
        Manually make a DELETE request.

        :param str path: relative url of the request (e.g. `/users/username`)
        :param auth.Authentication auth: authentication object
        :param kwargs dict: Extra arguments for the request, as supported by the
                            `requests <http://docs.python-requests.org/>`_ library.
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        return self._check_ok(self._delete(path, auth=auth, **kwargs))

    def _get(self, path, auth=None, **kwargs):
        if auth is not None:
            auth.update_kwargs(kwargs)
        try:
            return self._requestor.get(path, **kwargs)
        except requests.RequestException as exc:
            raise NetworkFailure(exc)

    def get(self, path, auth=None, **kwargs):
        """
        Manually make a GET request.

        :param str path: relative url of the request (e.g. `/users/username`)
        :param auth.Authentication auth: authentication object
        :param kwargs dict: Extra arguments for the request, as supported by the
                            `requests <http://docs.python-requests.org/>`_ library.
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        return self._check_ok(self._get(path, auth=auth, **kwargs))

    def _patch(self, path, auth=None, **kwargs):
        if auth is not None:
            auth.update_kwargs(kwargs)
        try:
            return self._requestor.patch(path, **kwargs)
        except requests.RequestException as exc:
            raise NetworkFailure(exc)

    def patch(self, path, auth=None, **kwargs):
        """
        Manually make a PATCH request.

        :param str path: relative url of the request (e.g. `/users/username`)
        :param auth.Authentication auth: authentication object
        :param kwargs dict: Extra arguments for the request, as supported by the
                            `requests <http://docs.python-requests.org/>`_ library.
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        return self._check_ok(self._patch(path, auth=auth, **kwargs))

    def _post(self, path, auth=None, **kwargs):
        if auth is not None:
            auth.update_kwargs(kwargs)
        try:
            return self._requestor.post(path, **kwargs)
        except requests.RequestException as exc:
            raise NetworkFailure(exc)

    def post(self, path, auth=None, **kwargs):
        """
        Manually make a POST request.

        :param str path: relative url of the request (e.g. `/users/username`)
        :param auth.Authentication auth: authentication object
        :param kwargs dict: Extra arguments for the request, as supported by the
                            `requests <http://docs.python-requests.org/>`_ library.
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        return self._check_ok(self._post(path, auth=auth, **kwargs))

    def _put(self, path, auth=None, **kwargs):
        if auth is not None:
            auth.update_kwargs(kwargs)
        try:
            return self._requestor.put(path, **kwargs)
        except requests.RequestException as exc:
            raise NetworkFailure(exc)

    def put(self, path, auth=None, **kwargs):
        """
        Manually make a PUT request.

        :param str path: relative url of the request (e.g. `/users/username`)
        :param auth.Authentication auth: authentication object
        :param kwargs dict: Extra arguments for the request, as supported by the
                            `requests <http://docs.python-requests.org/>`_ library.
        :raises NetworkFailure: if there is an error communicating with the server
        :raises ApiFailure: if the request cannot be serviced
        """
        return self._check_ok(self._put(path, auth=auth, **kwargs))

    @staticmethod
    def _check_ok(response):
        """
        Raise exception if response is non-OK, otherwise return response
        """
        if not response.ok:
            GiteaApi._fail(response)
        return response

    @staticmethod
    def _fail(response):
        """
        Raise an ApiFailure pertaining to the given response
        """
        message = "Status code: {}-{}, url: {}".format(response.status_code, response.reason, response.url)
        try:
            message += ", message:{}".format(response.json()["message"])
        except (ValueError, KeyError):
            pass
        raise ApiFailure(message, response.status_code)


class ApiFailure(Exception):
    """
    Raised to signal a failed request
    """

    def __init__(self, message, status_code):
        self._message = message
        self._status_code = status_code

    def __str__(self):
        return self._message

    @property
    def message(self):
        """
        An error message explaining why the request failed.

        :type: str
        """
        return self._message

    @property
    def status_code(self):
        """
        The HTTP status code of the response to the failed request

        :type: int
        """

        return self._status_code


class NetworkFailure(Exception):
    """
    Raised to signal a network-level failure
    """

    def __init__(self, cause=None):
        self._cause = cause

    def __str__(self):
        return "Caused by: {}".format(str(self._cause))

    @property
    def cause(self):
        """
        The exception causing the network failure

        :type: Exception
        """
        return self._cause



"""
Various HTTP utilities
"""

import requests
from future.moves.urllib.parse import urljoin


class RelativeHttpRequestor(object):
    """
    A thin wrapper around the requests module that allows for endpoint paths
    to be given relative to a fixed base URL
    """

    def __init__(self, base_url, session=None):
        self.base_url = base_url
        self.session = session or requests.Session()

    def absolute_url(self, relative_path):
        """
        :param str relative_path: relative URL
        :return: absolute URL of relative_path, relative to this object's base URL
        :rtype: str
        """
        return append_url(self.base_url, relative_path)

    # The below methods are identical to the corresponding functions in requests module,
    # except that they expect relative paths

    def delete(self, relative_path, **kwargs):
        return self.session.delete(self.absolute_url(relative_path), **kwargs)

    def get(self, relative_path, params=None, **kwargs):
        return self.session.get(self.absolute_url(relative_path), params=params, **kwargs)

    def options(self, relative_path, params=None, **kwargs):
        return self.session.options(self.absolute_url(relative_path), params=params, **kwargs)

    def patch(self, relative_path, data=None, **kwargs):
        return self.session.patch(self.absolute_url(relative_path), json=data, **kwargs)

    def post(self, relative_path, data=None, **kwargs):
        return self.session.post(self.absolute_url(relative_path), json=data, **kwargs)

    def put(self, relative_path, params=None, data=None, **kwargs):
        return self.session.put(self.absolute_url(relative_path), params=params, json=data, **kwargs)


def append_url(base_url, path):
    """
    Append path to base_url in a sensible way.
    """
    if base_url[-1] != "/":
        base_url += "/"
    if path[0] == "/":
        path = path[1:]
    return urljoin(base_url, path)



"""
Various classes for Gitea authentication
"""
from gitea_client.entities import json_get


class Authentication(object):
    """
    An "abstract" parent class. Represents a
    means of authenticating oneself to Gitea
    """

    def update_kwargs(self, kwargs):
        """
        Updates kwargs to include this object's authentication information
        :param dict kwargs: dictionary of keyword arguments to pass to a function from
            the requests module
        :return: Updated kwargs
        """
        raise NotImplementedError()  # must be implemented by subclasses


class Token(Authentication):
    """
    An immutable representation of a Gitea authentication token
    """

    def __init__(self, token, name=None):
        """
        :param str token: contents of Gitea authentication token
        """
        self._token = token
        self._name = name

    @staticmethod
    def from_json(parsed_json):
        name = json_get(parsed_json, "name")
        sha1 = json_get(parsed_json, "sha1")
        return Token(sha1, name)

    @property
    def name(self):
        """
        The name of the token

        :rtype: str
        """
        return self._name

    @property
    def token(self):
        """
        The contents of the Gitea authentication token

        :rtype: str
        """
        return self._token

    def update_kwargs(self, kwargs):
        if "params" in kwargs:
            kwargs["params"]["token"] = self._token
        else:
            kwargs["params"] = {"token": self._token}


class UsernamePassword(Authentication):
    """
    An immutable representation of a Gitea username/password combination
    """

    def __init__(self, username, password):
        """
        :param str username: Username for Gitea account
        :param str password: Password for Gitea account
        """
        self._username = username
        self._password = password

    @property
    def username(self):
        """
        Username for Gitea account

        :rtype: str
        """
        return self._username

    @property
    def password(self):
        """
        Password for Gitea account
        :rtype: str
        """
        return self._password

    def update_kwargs(self, kwargs):
        kwargs["auth"] = (self._username, self._password)



"""
Various classes representing update requests to Gitea
"""


class GiteaUserUpdate(object):
    """
    An immutable representation of a collection of Gitea user attributes to update.

    Instances should be created using the :class:`~GiteaUserUpdate.Builder` class.
    """

    def __init__(self, source_id, login_name, full_name, email, password, website,
                 location, active, admin, allow_git_hook, allow_import_local):
        self._source_id = source_id
        self._login_name = login_name
        self._full_name = full_name
        self._email = email
        self._password = password
        self._website = website
        self._location = location
        self._active = active
        self._admin = admin
        self._allow_git_hook = allow_git_hook
        self._allow_import_local = allow_import_local

    def as_dict(self):
        fields = {
            "source_id": self._source_id,
            "login_name": self._login_name,
            "full_name": self._full_name,
            "email": self._email,
            "password": self._password,
            "website": self._website,
            "location": self._location,
            "active": self._active,
            "admin": self._admin,
            "allow_git_hook": self._allow_git_hook,
            "allow_import_local": self._allow_import_local
        }
        return {k: v for (k, v) in fields.items() if v is not None}

    class Builder(object):
        def __init__(self, login_name, email):
            """
            :param str login_name: login name for authentication source
            :param str email: email address of user to update
            """
            self._source_id = None
            self._login_name = login_name
            self._full_name = None
            self._email = email
            self._password = None
            self._website = None
            self._location = None
            self._active = None
            self._admin = None
            self._allow_git_hook = None
            self._allow_import_local = None

        def set_source_id(self, source_id):
            """
            :param int source_id: Source id of authentication source

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._source_id = source_id
            return self

        def set_full_name(self, full_name):
            """
            :param str full_name: Updated full name for user

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._full_name = full_name
            return self

        def set_password(self, password):
            """
            :param str password: Updated password for user

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._password = password
            return self

        def set_website(self, website):
            """
            :param str website: Updated personal website for user

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._website = website
            return self

        def set_location(self, location):
            """
            :param str location: Updated location for user

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._location = location
            return self

        def set_active(self, active):
            """
            :param bool active: set true/false to signal that the updated user should be
                                activated/deactivated

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._active = active
            return self

        def set_admin(self, admin):
            """
            :param bool admin: whether the updated user should be admin

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._admin = admin
            return self

        def set_allow_git_hook(self, allow_git_hook):
            """
            :param bool allow_git_hook: whether the updated user should be allowed to use Git hooks

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._allow_git_hook = allow_git_hook
            return self

        def set_allow_import_local(self, allow_import_local):
            """

            :param bool allow_import_local: whether the updated user should be allowed to
                                            import local repositories

            :return: the updated builder
            :rtype: GiteaUserUpdate.Builder
            """
            self._allow_import_local = allow_import_local
            return self

        def build(self):
            """
            :return: A :class:`~GiteaUserUpdate` instance reflecting the changes added to the builder.
            :rtype: GiteaUserUpdate
            """
            return GiteaUserUpdate(
                source_id=self._source_id,
                login_name=self._login_name,
                full_name=self._full_name,
                email=self._email,
                password=self._password,
                website=self._website,
                location=self._location,
                active=self._active,
                admin=self._admin,
                allow_git_hook=self._allow_git_hook,
                allow_import_local=self._allow_import_local)


class GiteaHookUpdate(object):
    """
    An immutable representation of a collection of Gitea hook attributes to update.

    Instances should be created using the :class:`~GiteaHookUpdate.Builder` class.
    """

    def __init__(self, events, config, active):
        self._events = events
        self._config = config
        self._active = active

    def as_dict(self):
        fields = {
            "events": self._events,
            "config": self._config,
            "active": self._active,
        }
        return {k: v for (k, v) in fields.items() if v is not None}

    class Builder(object):
        def __init__(self):
            self._events = None
            self._config = None
            self._active = None

        def set_events(self, events):
            """
            :param List[str] events:
            :return: the updated builder
            :rtype: GiteaHookUpdate.Builder
            """
            self._events = events
            return self

        def set_config(self, config):
            """
            :param dict config:
            :return: the updated builder
            :rtype: GiteaHookUpdate.Builder
            """
            self._config = config
            return self

        def set_active(self, active):
            """
            :param bool active:
            :return: the updated builder
            :rtype: GiteaHookUpdate.Builder
            """
            self._active = active
            return self

        def build(self):
            """
            :return: A :class:`~GiteaHookUpdate` instance reflecting the changes added to the builder.
            :rtype: GiteaHookUpdate
            """
            return GiteaHookUpdate(
                events=self._events,
                config=self._config,
                active=self._active)



import unittest

import gitea_client._implementation.http_utils as http_utils


class HttpUtilsTest(unittest.TestCase):
    def test_append_url1(self):
        base = "https://www.google.com/"
        path = "/images/"
        self.assertEqual(http_utils.append_url(base, path), "https://www.google.com/images/")

    def test_append_url2(self):
        base = "http://www.hello.world.net"
        path = "great/api/v1"
        self.assertEqual(http_utils.append_url(base, path), "http://www.hello.world.net/great/api/v1")

    def test_append_url3(self):
        base = "https://hello.org/dir1/dir2"
        path = "/dir3/file.txt"
        self.assertEqual(http_utils.append_url(base, path),
                         "https://hello.org/dir1/dir2/dir3/file.txt")

    def test_absolute1(self):
        base = "https://www.google.com/"
        path = "/images/"
        requestor = http_utils.RelativeHttpRequestor(base)
        self.assertEqual(requestor.absolute_url(path), "https://www.google.com/images/")

    def test_absolute2(self):
        base = "http://www.hello.world.net"
        path = "great/api/v1"
        requestor = http_utils.RelativeHttpRequestor(base)
        self.assertEqual(requestor.absolute_url(path), "http://www.hello.world.net/great/api/v1")

    def test_absolute3(self):
        base = "https://hello.org/dir1/dir2"
        path = "/dir3/file.txt"
        requestor = http_utils.RelativeHttpRequestor(base)
        self.assertEqual(requestor.absolute_url(path),
                         "https://hello.org/dir1/dir2/dir3/file.txt")


if __name__ == "__main__":
    unittest.main()



"""
Various immutable classes that represent Gitea entities.
"""

from collections import OrderedDict

import attr


def json_get(parsed_json, key):
    """
    Retrieves the key from a parsed_json dictionary, or raises an exception if the
    key is not present
    """
    if key not in parsed_json:
        raise ValueError("JSON does not contain a {} field".format(key))
    return parsed_json[key]


@attr.s
class GiteaEntity(object):
    json = attr.ib()

    @classmethod
    def from_json(cls, parsed_json):
        # with introspection, get arguments of the constructor
        parsed_json['json'] = parsed_json.copy()
        params = cls.__attrs_attrs__
        args = []
        kwargs = OrderedDict()
        for param in params:
            param_name = param.name.lstrip('_')
            # if not a keyword argument
            if param.default == attr.NOTHING:
                args.append(json_get(parsed_json, param_name))
            # if it's a keyword argument
            else:
                kwargs[param_name] = parsed_json.get(param_name, None)
        o = cls(*args, **kwargs)
        return o


@attr.s(frozen=True)
class GiteaUser(GiteaEntity):
    """
     An immutable representation of a Gitea user
    """

    #: The user's id
    #:
    #: :type: int
    id = attr.ib()

    #: The user's id
    #:
    #: :type: int
    #:
    #: .. deprecated:: 1.1
    #:    Use :data:`id` instead
    user_id = property(lambda self: self.id)

    #: The user's username
    #:
    #: :type: str
    username = attr.ib()

    #: The user's full name
    #:
    #: :type: str
    full_name = attr.ib()

    #: The user's email address. Can be empty as a result of invalid authentication
    #:
    #: :type: str
    email = attr.ib(default=None)

    #: The user's avatar URL
    #:
    #: :type: str
    avatar_url = attr.ib(default=None)


@attr.s(frozen=True)
class GiteaRepo(GiteaEntity):
    """
    An immutable representation of a Gitea repository
    """

    #: The repository's id
    #:
    #: :type: int
    id = attr.ib()

    #: The repository's id
    #:
    #: :type: int
    #:
    #: .. deprecated:: 1.1
    #:    Use :data:`id` instead
    repo_id = property(lambda self: self.id)

    #: The owner of the repository
    #:
    #: :type: :class:`~GiteaUser`
    owner = attr.ib(converter=lambda parsed_json: GiteaUser.from_json(parsed_json))

    #: The name of the repository
    #:
    #: :type: str
    name = attr.ib()

    #: The full name of the repository
    #:
    #: :type: str
    full_name = attr.ib()

    #: Whether the repository is private
    #:
    #: :type: bool
    private = attr.ib()

    #: Whether the repository is a fork
    #:
    #: :type: bool
    fork = attr.ib()

    #: The name of the default branch
    #:
    #: :type: str
    default_branch = attr.ib()

    _html_url = attr.ib()
    _ssh_url = attr.ib()
    _clone_url = attr.ib()

    @property
    def urls(self):
        """
        URLs of the repository

        :type: :class:`~GiteaRepo.Urls`
        """
        return GiteaRepo.Urls(self._html_url, self._clone_url, self._ssh_url)

    #: Permissions for the repository
    #:
    #: :type: :class:`~GiteaRepo.Permissions`
    permissions = attr.ib(converter=lambda data: GiteaRepo.Permissions.from_json(data))

    #: Gets the repository's parent, when a fork
    #:
    #: :type: :class:`~GiteaRepo`
    parent = attr.ib(converter=lambda data: GiteaRepo.from_json(data) if data else None, default=None)

    #: The description of the repository
    #:
    #: :type: str
    description = attr.ib(default=None)

    #: Whether the repository is empty
    #:
    #: :type: bool
    empty = attr.ib(default=None)

    #: Size of the repository in kilobytes
    #:
    #: :type: int
    size = attr.ib(default=None)

    @attr.s(frozen=True)
    class Urls(object):
        #: URL for the repository's webpage
        #:
        #: :type: str
        html_url = attr.ib()

        #: URL for cloning the repository (via HTTP)
        #:
        #: :type: str
        clone_url = attr.ib()

        #: URL for cloning the repository via SSH
        #:
        #: :type: str
        ssh_url = attr.ib()

    @attr.s(frozen=True)
    class Permissions(GiteaEntity):
        #: Whether the user that requested this repository has admin permissions
        #:
        #: :type: bool
        admin = attr.ib(default=False)

        #: Whether the user that requested this repository has push permissions
        #:
        #: :type: bool
        push = attr.ib(default=False)

        #: Whether the user that requested this repository has pull permissions
        #:
        #: :type: bool
        pull = attr.ib(default=False)

    @attr.s(frozen=True)
    class Hook(GiteaEntity):
        #: The hook's id number
        #:
        #: :type: int
        id = attr.ib()

        #: The hook's id number
        #:
        #: :type: int
        #:
        #: .. deprecated:: 1.1
        #:    Use :data:`id` instead
        hook_id = property(lambda self: self.id)

        #: The hook's type (gitea, slack, etc.)
        #:
        #: :type: str
        type = attr.ib()

        #: The hook's type (gitea, slack, etc.)
        #:
        #: :type: str
        #:
        #: .. deprecated:: 1.1
        #:    Use :data:`type` instead
        hook_type = property(lambda self: self.type)

        #: The events that fire the hook
        #:
        #: :type: List[str]
        events = attr.ib()

        #: Whether the hook is active
        #:
        #: :type: bool
        active = attr.ib()

        #: Config of the hook. Possible keys include ``"content_type"``, ``"url"``, ``"secret"``
        #:
        #: :type: dict
        config = attr.ib()

    @attr.s(frozen=True)
    class DeployKey(GiteaEntity):
        #: The key's id number
        #:
        #: :type: int
        id = attr.ib()

        #: The key's id number
        #:
        #: :type: int
        #:
        #: .. deprecated:: 1.1
        #:    Use :data:`id` instead
        key_id = property(lambda self: self.id)

        #: The content of the key
        #:
        #: :type: str
        key = attr.ib()

        #: URL where the key can be found
        #:
        #: :type: str
        url = attr.ib()

        #: The name of the key
        #:
        #: :type: str
        title = attr.ib()

        #: Creation date of the key
        #:
        #: :type: str
        created_at = attr.ib()

        #: Whether key is read-only
        #:
        #: :type: bool
        read_only = attr.ib()


@attr.s(frozen=True)
class GiteaBranch(GiteaEntity):
    """
    An immutable representation of a Gitea branch
    """

    #: The branch's name
    #:
    #: :type: str
    name = attr.ib()

    #: The HEAD commit of the branch
    #:
    #: :type: :class:`~GiteaCommit`
    commit = attr.ib(converter=lambda parsed_json: GiteaCommit.from_json(parsed_json))


@attr.s(frozen=True)
class GiteaCommit(GiteaEntity):
    """
    An immutable representation of a Gitea commit
    """

    #: The commit's id
    #:
    #: :type: str
    id = attr.ib()

    #: The commit's message
    #:
    #: :type: str
    message = attr.ib()

    #: The commit's url
    #:
    #: :type: str
    url = attr.ib()

    #: The commit's timestamp
    #:
    #: :type: str
    timestamp = attr.ib()


@attr.s(frozen=True)
class GiteaOrg(GiteaEntity):
    """
     An immutable representation of a Gitea organization
    """

    #: The organization's id
    #:
    #: :type: int
    id = attr.ib()

    #: The organization's id
    #:
    #: :type: int
    #:
    #: .. deprecated:: 1.1
    #:    Use :data:`id` instead
    org_id = property(lambda self: self.id)

    #: Organization's username
    #:
    #: :type: str
    username = attr.ib()

    #: Organization's full name
    #:
    #: :type: str
    full_name = attr.ib()

    #: Organization's avatar URL
    #:
    #: :type: str
    avatar_url = attr.ib()

    #: Organization's description
    #:
    #: :type: str
    description = attr.ib()

    #: Organization's website address
    #:
    #: :type: str
    website = attr.ib()

    #: Organization's location
    #:
    #: :type: str
    location = attr.ib()


@attr.s(frozen=True)
class GiteaTeam(GiteaEntity):
    """
    An immutable representation of a Gitea organization team
    """

    #: Team's id
    #:
    #: :type: int
    id = attr.ib()

    #: Team's id
    #:
    #: :type: int
    #:
    #: .. deprecated:: 1.1
    #:    Use :data:`id` instead
    team_id = property(lambda self: self.id)

    #: Team name
    #:
    #: :type: str
    name = attr.ib()

    #: Description of the team
    #:
    #: :type: str
    description = attr.ib()

    #: Team permission, one of `"read"`, `"write"`, `"admin"`, `"owner"` or `"none"`
    #:
    #: :type: str
    permission = attr.ib()



from codecs import open
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.rst"), "r") as f:
    long_description = f.read()

setup(
    name="gitea_client",
    version="1.0.7",
    description="A python library for interacting with a Gitea server",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/unfoldingWord-dev/python-gitea-client",
    author="unfoldingWord",
    author_email="dev@unfoldingword.org",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    keywords=["gitea", "gogs", "http", "client"],
    packages=find_packages(),
    install_requires=["future", "requests", "attrs"],
    test_suite="tests"
)
